############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import sha, base64, time

import ZSI
from ZSI.wstools.Namespaces import DSIG, OASIS, GLOBUS, WSA200403
from pyGridWare.security import VerifyError
from pyGridWare.security.gss.GssSecureConversation import GssSecurityContext
from pyGridWare.security.gss.GssSecureMessage import GSIAuthentication
from pyGridWare.utility.dom import  DomletteElementProxy
from pyGridWare.generated.types import GetNSContainer
import pyGridWare.generated.types.SecureConversation_services_types as _AUTH
_xmldsig_class = GetNSContainer(_AUTH,"http://www.w3.org/2000/09/xmldsig#")
_wssecurity_secext_class = GetNSContainer(_AUTH, "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd")
from pyGridWare.security.base.BaseSignatureHandler import BaseSignatureHandler
from pyGridWare.security.base.BaseSignatureHandler import SignatureHandlerError

class VerifyError(SignatureHandlerError):
    pass

class Holder: 
    pass

class Signature:
    typecode = _xmldsig_class.Signature_Dec()


class BST(str):
    typecode = _wssecurity_secext_class.BinarySecurityToken_Dec()
    
class TimeStamp(ZSI.TC.Struct):
    attribute_typecode_dict = {(OASIS.UTILITY,"Id"): ZSI.TC.String() }
    def __init__(self):
        pname = (OASIS.UTILITY,"TimeStamp")
        tclist = [ZSI.TCtimes.gDateTime(pname=(OASIS.UTILITY,"Created"),aname="_Created",minOccurs=1,maxOccurs=1,nillable=0),ZSI.TCtimes.gDateTime(pname=(OASIS.UTILITY,"Expires"),aname="_Expires",minOccurs=1,maxOccurs=1,nillable=0)]
        ZSI.TCcompound.Struct.__init__(self, self.__class__, tclist,pname)

class KeyInfo:
    typecode =_xmldsig_class.KeyInfo_Dec


class SecurityReference:
    typecode = _wssecurity_secext_class.SecurityTokenReference_Dec()


class Reference:
    typecode = _wssecurity_secext_class.Reference_Dec()
    

class GssSignatureHandler(BaseSignatureHandler):
    """
    This class is responsible for signing and verifying messages, using
    GssSecureConversation or GssSecureMessage protocols.
    """
    
    def __init__(self):
        self._can_algo = DSIG.C14N
        self._sig_algo = GLOBUS.SIG
        self._dig_algo = DSIG.DIGEST_SHA1
        self._contextID = None
        self._expire_time = 300
        self._setProtocol = False
        self._counterKey = None
        
    def setContextID(self, KeyName):
        self._contextID =  KeyName

    def getContextID(self):
        return self._contextID

    def protocolSet(self):
        return self._setProtocol

    def verify(self, ps):
        """
        The algorithm to verify:

        Verify the digests
        
        Get and canonicalize the signedInfo node

        Get the value of the SignatureValue node

        base64 decode the SignatureValue

        verify the SignatureValue 
        """
        element = DomletteElementProxy(None, message=ps.header)
        element.setContext(processorNss={'ds':DSIG.BASE, 'wsu':OASIS.UTILITY})

        #This is how the a Server side service figures out which protocol to use
        if not self.protocolSet():
            signatureMethodNodes = element.evaluate(expression='//ds:SignatureMethod')
            signatureMethodNode = signatureMethodNodes[0]
            signatureAlgorithm = signatureMethodNode.getAttributeValue(None, "Algorithm")
            contextID = None
            if signatureAlgorithm == GLOBUS.SIG:
                keyNameNodes = element.evaluate(expression='//ds:KeyName')
                keyNameNode = keyNameNodes[0]
                contextID = keyNameNode.getValue()
            self.setProtocol(signatureAlgorithm, contextID)

        useSecureMessage = self.useSecureMessage()
        
        referenceNodes = element.evaluate(expression='//ds:Reference')
        for refNode in referenceNodes:
            if useSecureMessage:
                transformsNode = refNode.getElement(DSIG.BASE, "Transforms")
                #There could be more than one transform or no elements according to the schema 
                if transformsNode is not None:
                    transformNode = transformsNode.getElement(DSIG.BASE, "Transform")
                    algorithm = transformNode.getAttributeValue(None,"Algorithm")
                else:
                    algorithm = self._can_algo
            else:
                algorithm = self._can_algo
            uri = refNode.getAttributeValue(None, "URI")
            digestValue = self.getDigestValueFromDom(element, uri, algorithm)
            digestValue64 = base64.encodestring(digestValue)
            digestValueStripped = digestValue64.strip()
            digestNode = refNode.getElement(DSIG.BASE,"DigestValue")
            receivedDigestValue = digestNode.getValue()
            receivedDigestValueStripped = receivedDigestValue.strip()
            if digestValueStripped != receivedDigestValueStripped:
                raise VerifyError("DigestValues do not match")
                   
        signedInfoNodes = element.evaluate(expression='//ds:SignedInfo')
        if len(signedInfoNodes) != 1:
             raise VerifyError("more than one uri found with uri: " + str(len(signedInfoNodes)))

        signedInfoNode = signedInfoNodes[0]
        canNode = signedInfoNode.getElement(DSIG.BASE, "CanonicalizationMethod")
        algorithm =  canNode.getAttributeValue(None, "Algorithm")
        signedInfoStr = signedInfoNode.canonicalize(algorithm)
        signedInfoHash = sha.sha(signedInfoStr).digest()
        signatureValueNodes = element.evaluate(expression='//ds:SignatureValue')
        if len(signatureValueNodes) != 1:
            raise VerifyError("more than one uri found with uri: " + str(len(signatureValueNodes)))
        signatureValueNode = signatureValueNodes[0]
        signatureStr = signatureValueNode.getValue()
        
        keyInfoNodes = element.evaluate(expression='//ds:KeyInfo')
        keyInfoNode = keyInfoNodes[0]
        if useSecureMessage:
            secReferenceNode = keyInfoNode.getElement(OASIS.WSSE, "SecurityTokenReference")
            referenceNode = secReferenceNode.getElement(OASIS.WSSE, "Reference")
            keyUri = referenceNode.getAttributeValue(None, "URI")
            bst = self.getSecurityToken(element, keyUri)
            auth = GSIAuthentication()
       
       
        if useSecureMessage:
            verified = auth.verify(base64.decodestring(bst), signedInfoHash, base64.decodestring(signatureStr))
        else:
            home = GssSecurityContext.getResourceHome()
            secContext = home.get(contextID)
            verified = secContext.verifyMIC(signedInfoStr, base64.decodestring(signatureStr))
        if not verified:
            raise VerifyError("Signatures do not match")
       
       
    def sign(self, sw):
        """
        Sign the message whether we are using
        GssSecureConversation or GssSecureMesage

        General algorithm
        - Create the authentication information
        elements (ie BinarySecurityToken)
        and security elements(ie timestamp)
        
        - Create digests for anything in the
        body of the Message

        - Fill in the SignedInfoNode, create elements
        for digests of elements in the header
        leave the content of the element empty. Create
        a SignatureValue element also leavethe content
        of that element empty.

        - serialize the header

        - Calculate the digests for elements in the
        header(we had to wait until the header was
        serialized to do this) Put the digests as
        the content for those empty digestValue elements
        that we had created

        - Calculate the SignatureValue and place it in
        the content of the SignatureValue element
        
        """

        context_id = self._contextID

        useSecureMessage = self.useSecureMessage()
        self.signHeaderElements(sw)
        
        idUriList = []
        if useSecureMessage:
            auth = GSIAuthentication()
            pkiPath = auth.getPkiPath()
            pkiPath64 = base64.encodestring(pkiPath)
            binarySecurityToken = BST(pkiPath64)
            #move this GssSecureMessage
            tokenRefString = self.getTokenRefString()
            binarySecurityToken._attrs = {(OASIS.UTILITY,'Id'): "CertId-" +tokenRefString,
                        'ValueType':"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-x509-token-profile-1.0#X509PKIPathv1",
                        'EncodingType':"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary"}
            binarySecurityToken.typecode.attribute_typecode_dict.update({(OASIS.UTILITY,'Id'):ZSI.TC.String(),
                                                                         'ValueType': ZSI.TC.String(),
                                                                         'EncodingType':ZSI.TC.String()})

            """
            There are some problems with the generated code that I need to work out
            then I can do this
            timeStamp = _AUTH.ns1.Timestamp_Dec()
            timeStamp.attribute_typecode_dict = {(OASIS.UTILITY,"Id"): ZSI.TC.String() }
            timeStamp._Created = self.makeTimeStampString(time.gmtime())
            expires = time.time() + self._expire_time
            timeStamp._Expires = self.makeTimeStampString(time.gmtime(expires))
            """
            timeStamp = TimeStamp()
            timeStamp._Created = time.gmtime()
            expires = time.time() + self._expire_time
            timeStamp._Expires = time.gmtime(expires)
            timeStamp._attrs = {(OASIS.UTILITY,"Id"): "timestamp"}
            idUriList.append("#timestamp")

         
        typecode = _wssecurity_secext_class.Security_Dec()
        pyobj = Holder()
        pyobj._any = []
        if useSecureMessage:
            pyobj._any.append(timeStamp)
            pyobj._any.append(binarySecurityToken)   
             
                                 
        sig = Signature()
        sig._SignedInfo = Holder()
        sig._SignedInfo._CanonicalizationMethod = Holder()
        sig._SignedInfo._CanonicalizationMethod._attrs = {'Algorithm':self._can_algo}
        sig._SignedInfo._SignatureMethod = Holder()
        sig._SignedInfo._SignatureMethod._attrs = {'Algorithm':self._sig_algo}
        sig._SignedInfo._Reference = []
        sig._SignatureValue = ''
        
        if useSecureMessage:
            sig._KeyInfo = KeyInfo()
            secRef = SecurityReference()
            secRef._any = Reference()
            secRef._any._attrs = {}
            secRef._any._attrs['URI'] = "#CertId-" + tokenRefString
            sig._KeyInfo._any = secRef 
        else:
            home = GssSecurityContext.getResourceHome()
            secContext = home.get(context_id)
            sig._KeyInfo = Holder()
            sig._KeyInfo._KeyName = context_id
        
        processorNss={'wsu':OASIS.UTILITY, 'ds':DSIG.BASE}
        sw.dom.setContext(processorNss=processorNss)
        nodes = sw.dom.evaluate(expression='//*[@wsu:Id]')

        #Calculate the digests in the body
        for node in nodes:
            URI = '#%s' %node.getAttributeValue(OASIS.UTILITY, 'Id')
            ref = Holder()
            ref._attrs = {'URI': URI}
            ref._DigestMethod = Holder()
            ref._DigestMethod._attrs = {'Algorithm':self._dig_algo}
            ref._DigestValue = self.getDigestValueFromSoapWriter(sw, URI)
            sig._SignedInfo._Reference.append(ref)

        #These haven't been serialized yet, the id's to take
        #digests of in the security header
        for idUri in idUriList:
            ref = Holder()
            ref._attrs = {'URI': idUri}
            ref._DigestMethod = Holder()
            ref._DigestMethod._attrs = {'Algorithm':self._dig_algo}
            ref._DigestValue = ''
            sig._SignedInfo._Reference.append(ref)

        pyobj._any.append(sig)
        sw.serialize_header(pyobj, typecode)

        # Reset Context after append on DOM
        sw.dom.setContext(processorNss=processorNss)
        for URI in idUriList:
            digestValue = self.getDigestValueFromSoapWriter(sw, URI)
            referenceNodes = sw.dom.evaluate(expression='//ds:Reference')
            
            for referenceNode in referenceNodes:
                refUri = referenceNode.getAttributeValue(None, "URI")
                if refUri == URI:
                    digestValueNode = referenceNode.getElement(DSIG.BASE, "DigestValue")
                    digestValueNode.createAppendTextNode(base64.encodestring(digestValue))
            
        # Reset Context after append on DOM
        sw.dom.setContext(processorNss=processorNss)
        nodes = sw.dom.evaluate(expression='//ds:SignedInfo')
        signedInfo = nodes[0]
        signedInfoStr = signedInfo.canonicalize()
        nodes = sw.dom.evaluate(expression='//ds:SignatureValue')
        signatureValueNode = nodes[0]
        if useSecureMessage:
            hashedSignedInfoStr = sha.sha(signedInfoStr).digest()
            signatureValue = base64.encodestring(auth.sign(hashedSignedInfoStr))
        else:
            signatureValue = base64.encodestring(secContext.getMIC(signedInfoStr))
        signatureValueNode.createAppendTextNode(signatureValue)


    def getSecurityToken(self, element, uri):
        uriExpString = '//*[@wsu:Id="' + uri[1:] +'"]'
        if not element.isContextInitialized():
            raise SignatureHandlerError('ElementProxy has no context')
        uriNodes = element.evaluate(expression=uriExpString)
        uriNode = uriNodes[0]
        bst = uriNode.getValue()
        return bst
    
    def setSecureMessage(self):
        """
        The signature handler will sign and verify messages according
        to the GssSecureMessage protocol
        """
        self.setProtocol(DSIG.SIG_RSA_SHA1)

    def setSecureConversation(self, contextID):
        """
        The signature handler will sign and verify messages according
        to the GssSecureConversation protocol
        """
        self.setProtocol(GLOBUS.SIG, contextID)
    
    def setProtocol(self, uri, contextID=None):
        """
        sets the protocol uri, and the contextID

        **Raises**
        A SignatureHandlerError is raised if the signature algorithm is not supported
        """
        if uri == DSIG.SIG_RSA_SHA1:
            if contextID is not None:
                raise SignatureHandlerError("contextID must be None if you are using GssSecureMessage")
        elif uri == GLOBUS.SIG:
            if contextID is None:
                raise SignatureHandlerError("contextID must not be None if you are using GssSecureConveration") 
            
        else:
            raise SignatureHandlerError("Unsupported signature algorithm: " + uri)

        self._sig_algo = uri
        self._contextID = contextID
        self._protocolSet = True

    def useSecureMessage(self):
        """
        returns True if we are using GssSecureMessage and otherwise it returns False
        """
        return self._sig_algo == DSIG.SIG_RSA_SHA1
        

    def makeTimeStampString(self, timeVal):
        """
        makes a timestamp string according to the OASIS schema

        **Arguments**

        - *timeVal* is a tuple of a the time to construct the
        string
        """
        timeString = self.normalizeTimeCharacter(timeVal[0]) + '-'+ \
        self.normalizeTimeCharacter(timeVal[1]) + '-' \
        + self.normalizeTimeCharacter(timeVal[2]) + '-' + 'T' + \
        self.normalizeTimeCharacter(timeVal[3]) +':'+ \
        self.normalizeTimeCharacter(timeVal[4]) + \
        ':' +  self.normalizeTimeCharacter(timeVal[5]) + 'Z'
        return timeString

    def normalizeTimeCharacter(self, timeVal):
        strVal = str(timeVal)
        if len(strVal) == 1:
            strVal = '0' + strVal

        return strVal
