###########################################################################
# Monte M. Goode, LBNL
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################

# $Id: resource.py,v 1.9 2004/11/24 00:31:56 boverhof Exp $

import os,sys
from twisted.web.server import NOT_DONE_YET
from twisted.web.resource import Resource
from twisted.python import log, failure

from ZSI import EvaluateException, ParseException
from ZSI.parse import ParsedSoap
from ZSI.writer import SoapWriter
from ZSI.fault import FaultFromException, FaultFromZSIException
from ZSI.address import Address
from ZSI.ServiceContainer import _Dispatch, WSActionException, WSResourceService,\
    PostNotSpecified, SOAPActionNotSpecified, WSActionNotSpecified

from pyGridWare.utility.dom import DomletteReader, DomletteElementProxy


class SimpleWSResource(WSResourceService):
    '''
    class variables:
        encoding  
        wsAction -- Must override to set output Action values.
        soapAction -- Must override to set input Action values.
    '''
    encoding = "UTF-8"
    def __init__(self, post=None):
        '''
        post -- POST value
        '''
        assert isinstance(self.soapAction, dict), "soapAction must be a dict"
        assert isinstance(self.wsAction, dict), "wsAction must be a dict"
        WSResourceService.__init__(self, post)

    def do_POST(self):
        '''The POST command.  This is called by HTTPServer, not twisted.
        action -- SOAPAction(HTTP header) or wsa:Action(SOAP:Header)
        '''
        soapAction = self.headers.getheader('SOAPAction')
        post = self.path
        if not post:
            raise PostNotSpecified, 'HTTP POST not specified in request'
        if not soapAction:
            raise SOAPActionNotSpecified, 'SOAP Action not specified in request'
        soapAction = soapAction.strip('\'"')
        post = post.strip('\'"')
        try:
            ct = self.headers['content-type']
            if ct.startswith('multipart/'):
                cid = resolvers.MIMEResolver(ct, self.rfile)
                xml = cid.GetSOAPPart()
                ps = ParsedSoap(xml, resolver=cid.Resolve, readerclass=DomletteReader)
            else:
                length = int(self.headers['content-length'])
                ps = ParsedSoap(self.rfile.read(length), readerclass=DomletteReader)
        except ParseException, e:
            #self.send_fault(FaultFromZSIException(e))
            raise
        except Exception, e:
            # Faulted while processing; assume it's in the header.
            self.send_fault(FaultFromException(e, 1, sys.exc_info()[2]))
        else:
            _Dispatch(ps, self.server, self.send_xml, self.send_fault,
                post=post, action=soapAction)

    
class TwistedWSResource(Resource):
    '''
    class variables:
        encoding  
        soapAction -- Must override to set input Action values.
           Dictionary of soapAction keys, and operation name values.
           These are specified in the WSDL soap bindings. There must be a 
           class method matching the operation name value.  If WS-Action is
           used the keys are WS-Action request values, according to the spec
           if soapAction and WS-Action is specified they must be equal.
           
        wsAction -- Must override to set output Action values.
           Dictionary of operation name keys and WS-Action 
           response values.  These values are specified by the portType.
    '''
    encoding = "UTF-8"
    wsAction = None
    soapAction = None

    def __init__(self, post):
        '''
        post -- POST value, also have access via Resource.
        '''
        assert isinstance(self.soapAction, dict), "soapAction must be a dict"
        assert isinstance(self.wsAction, dict), "wsAction must be a dict"
        self.post = post
        self._url = None
        Resource.__init__(self)

    def sign(self, sw):
        '''Internal Function, Override for security
        '''
        return

    def verify(self, ps):
        '''Internal Function, Override for security
        '''
        return

    def getServiceURL(self):
        return self._url

    def setServiceURL(self, url):
        self._url = url

    def getOperation(self, address):
        '''Returns a method of class.
        address -- ws-address 
        '''
        action = address.getAction()
        opName = self.getOperationName(action)
        return getattr(self, opName)

    def getOperationName(self, action):
        '''Returns operation name.
           action -- soapAction value
        '''
        if not self.soapAction.has_key(action):
            raise SOAPActionNotSpecified, 'soapAction dictionary missing key(%s)' %action
        return self.soapAction[action]

    def getResponseAction(self, action):
        '''Returns response WS-Action if available
           action -- request WS-Action value.
        '''
        opName = self.getOperationName(action)
        if not self.wsAction.has_key(opName):
            raise WSActionNotSpecified, 'wsAction dictionary missing key(%s)' %opName
        return self.wsAction[opName]

    def render_POST(self, request):
        """ Method called by twisted, not by HTTPServer.
        request -- twisted.web.server.Request
        """
        if request.path != self.post:
            raise Warning, 'The request path(%s) doesn\'t match the service post(%s)' %(request.path,self.post)
        http_headers = request.getAllHeaders()
        soap_action = http_headers.get('soapaction').strip('\'"')
        log.debug('BEGIN REQUEST %s' %soap_action)
        data = request.content.read()
        log.debug(data)
        try:
            ps = ParsedSoap(data, readerclass=DomletteReader)
        except ParseException, e:
            self._zsiException(request, e)
            return NOT_DONE_YET

        # WS-Address
        address = Address()
        netloc = request.getHeader('host')
        localURL = 'http://%s%s' %(netloc, request.path)
        log.debug('Service local URL: %s' %localURL)
        self.setServiceURL(localURL)
        try:
            address.parse(ps)
        except ParseException, e:
            # THROW A WS-Address Error
            self._zsiException(request, e)
            return NOT_DONE_YET

        action = address.getAction()
        if not action:
            log.err('No WS-Action specified in Request')
            e = WSActionException('No WS-Action specified in Request')
            self._zsiException(request, e)
            return NOT_DONE_YET

        if soap_action and soap_action != action:
            log.err('SOAP Action("%s") must match WS-Action("%s") if specified.' %(soap_action,action))
            e = WSActionException('SOAP Action("%s") must match WS-Action("%s") if specified.' %(soap_action,action))
            self._zsiException(request, e)
            return NOT_DONE_YET

        try:
            bound_method = self.getOperation(address)
        except SOAPActionNotSpecified, ex:
            log.err('Failed to locate service instance/bound method from WS-Address')
            self._zsiException(request, ex)
            return NOT_DONE_YET

        if not bound_method or not callable(bound_method):
            log.err('WS-Action("%s") must map to a callable method' %action)
            self._methodNotFound(request, ps)
            return NOT_DONE_YET

        try:
            result = bound_method(ps, address)
        except Exception, ex:
            #log.err('Failed to call method %s' %bound_method)
            #self._zsiException(request, ex)
            #return NOT_DONE_YET
            raise

        log.debug('Result %s' %str(result))
        # Verify if Signed
        self.verify(ps)

        # If No response just return.
        if result is None:
            log.debug('No Response specified')
            return NOT_DONE_YET

        sw = SoapWriter(outputclass=DomletteElementProxy)
        try:
            sw.serialize(result)
        except Exception, ex:
            e = FaultFromException(ex, 0, sys.exc_info()[2])
            self._zsiException(request, ex)
            return NOT_DONE_YET

        rsp_action = self.getResponseAction(action)
        addressRsp = Address(action=rsp_action)
        addressRsp.setResponseFromWSAddress(address, localURL)
        try:
            addressRsp.serialize(sw)
        except Exception, ex:
            log.err('Problem serializing WS-Address for Response')
            self._zsiException(request, e)
            return NOT_DONE_YET

        self.sign(sw)
        soapdata = str(sw)
        self._sendResponse(request, soapdata)
        return NOT_DONE_YET

    def _sendResponse(self, request, response, status=200):
        request.setResponseCode(status)

        if self.encoding is not None:
            mimeType = 'text/xml; charset="%s"' % self.encoding
        else:
            mimeType = "text/xml"

        log.debug(response)
            
        request.setHeader("Content-type", mimeType)
        request.setHeader("Content-length", str(len(response)))
        request.write(response)
        request.finish()

    # these are simply exception handling methods

    def _gotError(self, f, request, method):
        response = FaultFromException(f.value, 1).AsSOAP()

        log.err('XXX: gotError %s' % response)

        self._sendResponse(request, response, status=500)

    def _zsiException(self, request, e):
        response = FaultFromZSIException(e).AsSOAP()

        log.err('XXX: zsi exception: %s' % response)

        self._sendResponse(request, response, status=500)

    def _methodNotFound(self, request, ps):

        log.err('XXX: method not found')
        log.err('XXX: soapaction: %s' % request.received_headers['soapaction'])
        log.err('XXX: local name: %s' % ps.body_root.localName )
        
        response = FaultFromException(
            'Could not find exposed method - soapaction: %s, message: %s' \
            % (request.received_headers['soapaction'],
               ps.body_root.localName), 1 ).AsSOAP()

        self._sendResponse(request, response, status=500)

    def _typeError(self, request, method, result):
        response = FaultFromException(
            'TypeError: method %s improper return type %s' \
            % (method, result), 1 ).AsSOAP()

        log.err('XXX: typeError: %s' % response)

        self._sendResponse(request, response, status=500)

    def _typecodeNotFound(self, request, typecode):
        response = FaultFromException(
            'Could not find typecode %s' % typecode, 1 ).AsSOAP()

        log.err('XXX: typecode not found: %s' % response)

        self._sendResponse(request, response, status=500)



class WSResourceMetaClass(type):
    """Determines the Appropriate Base Class for a WSResource,
    either can be a Twisted Application, or a Stand-Alone ZSI
    Server.  Default to Twisted.
    """
    def __new__(cls, classname, bases, classdict):
        twisted = int(os.environ.get("USE_TWISTED", "1"))
        if twisted:
            bases = (TwistedWSResource,) + bases
        else:
            bases = (SimpleWSResource,) + bases
        for attr,item in classdict.items():
            # Sanity check just to make sure not screwing up
            # default constructor call.
            assert (attr!="__init__") or callable(item) is False, 'Must Not Override __init__'
        obj = type(classname, bases, classdict)
        return obj


class WSResource(object):
    '''Base class for all WS-Resource Services, chooses
    the correct implementation at runtime.  The two 
    implementations have different interfaces, but they
    are ONLY exposed to the Application Server.
    
      twisted    --> render_POST
      HTTPServer --> do_POST
    '''
    __metaclass__ = WSResourceMetaClass
    encoding = "UTF-8"



