############################################################################
# Monte M. Goode, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################

# contains text container classes for new generation generator

# $Id: containers.py,v 1.1.2.12 2004/11/30 21:50:47 boverhof Exp $

import re, sys
from cStringIO import StringIO
from utility import StringWriter, NamespaceAliasDict
from ZSI.wstools import XMLSchema
from ZSI.wstools.Namespaces import SCHEMA
from ZSI.wstools.WSDLTools import SoapBodyBinding
from ZSI.typeinterpreter import BaseTypeInterpreter
from ZSI.generate import WSISpec, WSInteropError

ID1 = '    '
ID2 = 2*ID1
ID3 = 3*ID1
ID4 = 4*ID1
ID5 = 5*ID1
ID6 = 6*ID1

CONTAINER_DEBUG = True

class ContainerError(Exception):
    pass

class ContainerBase:
    '''Base class for all Containers.
    '''
    def __init__(self):
        self.content = StringWriter('\n')
        self.setup   = False

        self.ns    = None
        self.nsad  = NamespaceAliasDict()

        if CONTAINER_DEBUG:
            self.write('# Container: <%s>\n' % self.__class__)

    def __str__(self):
        return self.getvalue()

    # - string content methods
    def mangle(self, s):
        '''class/variable name illegalities
        '''
        if s:
            return re.sub('[-./:?*~*# ]', '_', s)
        else:
            return None

    def write(self, s):
        self.content.write(s)

    def writeArray(self, a):
        self.content.write('\n'.join(a))

    def _setContent(self):
        '''override in subclasses.  formats the content in the desired way.
        '''
        raise NotImplementedError, 'abstract method not implemented'

    def getvalue(self):
        if not self.setup:
            self._setContent()
            self.setup = True
            
        return self.content.getvalue()

    # - namespace utility methods
    def getNSAlias(self):
        if self.ns:
            return self.nsad.getAlias(self.ns)
        else:
            raise ContainerError, 'no self.ns attr defined in %s' \
                  % self.__class__

    def getNSModuleName(self):
        if self.ns:
            return self.nsad.getModuleName(self.ns)
        else:
            raise ContainerError, 'no self.ns attr defined in %s' \
                  % self.__class__


# -- containers for services file components

class ServiceContainerBase(ContainerBase):
    def __init__(self):
        ContainerBase.__init__(self)
    
class ServiceHeaderContainer(ServiceContainerBase):
    def __init__(self):
        ServiceContainerBase.__init__(self)
        
        self.types = None
        self.extras = []

    def _setContent(self):
        
        if self.types:
            self.write('from %s import *' % self.types)
                
        imports = [
            '\nimport urlparse, types',
            'from ZSI.TCcompound import Struct',
            'from ZSI import client',
            'import ZSI'
            ]

        imports += self.extras

        self.writeArray(imports)

class ServiceInterfaceContainer(ServiceContainerBase):
    def __init__(self):
        ServiceContainerBase.__init__(self)
        
        self.serviceName = None
        self.portInfo = []

    def addPortInfo(self, ptName):
        self.portInfo.append(ptName)

    def _setContent(self):

        if not self.serviceName:
            raise ContainerError, 'no service name defined!'
        else:
            self.serviceName = self.mangle(self.serviceName)

        interface = [
            '# Interface',
            'class %sInterface:' % self.serviceName,
            ]

        for p in self.portInfo:
            pI = [
                '%sdef get%s(self, portAddress=None, **kw):' % (ID1, p),
                '%sraise NonImplementationError, "not implemented"' % ID2,
                ]

            interface += pI

        self.writeArray(interface)


class ServiceLocatorContainer(ServiceContainerBase):
    def __init__(self):
        ServiceContainerBase.__init__(self)
        
        self.serviceName = None
        self.portInfo = []
        self.locatorName = None
        self.portMethods = []

    def addPortInfo(self, ptName, bName, sAddress):
        self.portInfo.append( (ptName, bName, sAddress) )

    def getLocatorName(self):
        '''return class name of generated locator.
        '''
        return self.locatorName

    def getPortMethods(self):
        '''list of get port accessor methods of generated locator class.
        '''
        return self.portMethods

    def _setContent(self):
        if not self.serviceName:
            raise ContainerError, 'no service name defined!'
        else:
            self.serviceName = self.mangle(self.serviceName)

        self.locatorName = '%sLocator' %self.serviceName
        locator = [
            '# Locator',
            'class %s(%sInterface):' % (self.locatorName, self.serviceName),
            ]

        self.portMethods = []
        for p in self.portInfo:
            ptName = p[0]
            bName  = p[1]
            sAdd   = p[2]
            method = 'get%s' %ptName
            pI = [
                '%s%s_address = "%s"' % (ID1, ptName, sAdd),
                '%sdef get%sAddress(self):' % (ID1, ptName),
                '%sreturn %sLocator.%s_address' % (ID2,
                                                   self.serviceName,ptName),
                '%sdef %s(self, portAddress=None, **kw):' %(ID1, method),
                '%sreturn %sSOAP(portAddress or %sLocator.%s_address, **kw)' \
                % (ID2, bName, self.serviceName, ptName),
                ]
            self.portMethods.append(method)
            locator += pI

        self.writeArray(locator)


class ServiceOperationsClassContainer(ServiceContainerBase):
    def __init__(self):
        ServiceContainerBase.__init__(self)

        self.rProp = None
        self.wsA   = None
        self.bName = None

        self.operations = []
        
        self.reader = 'None'
        self.writer = 'None'

    def addOperation(self, op):
        self.operations.append(op)

    def _setContent(self):

        if self.wsA:
            ctorArgs = 'endPointReference=None, **kw'
            epr      = 'self.endPointReference = endPointReference'
        else:
            ctorArgs = '**kw'
            epr      = '# no ws-addressing'

        if self.rProp:
            rprop = 'kw["ResourceProperties"] = ("%s","%s")' % (self.rProp[0],
                                                                self.rProp[1])
        else:
            rprop = '# no resource properties'

        methods = [
            '# Methods',
            'class %sSOAP:' % self.bName,
            '%sdef __init__(self, addr, %s):' % (ID1, ctorArgs),
            '%snetloc = (urlparse.urlparse(addr)[1]).split(":") + [80,]' % ID2,
            '%sif not kw.has_key("host"):' % ID2,
            '%skw["host"] = netloc[0]' % ID3,
            '%sif not kw.has_key("port"):' % ID2,
            '%skw["port"] = int(netloc[1])' % ID3,
            '%sif not kw.has_key("url"):' % ID2,
            '%skw["url"] =  urlparse.urlparse(addr)[2]' % ID3,
            '%skw["readerclass"] = %s' % (ID2, self.reader),
            '%skw["writerclass"] = %s' % (ID2, self.writer),
            '%s%s' % (ID2, rprop),
            '%sself.binding = client.Binding(wsAddressing=%s, **kw)' \
            %(ID2, self.wsA),
            '%s%s' % (ID2,epr),
            ]

        for op in self.operations:
            methods += [ op.getvalue() ]

        self.writeArray(methods)

class ServiceOperationContainer(ServiceContainerBase):
    def __init__(self):
        ServiceContainerBase.__init__(self)

        self.name = None
        self.wsA  = None
        self.soapaction = None

        self.inputName  = None
        self.outputName = None
        
        self.inputSimpleType  = None
        self.outputSimpleType = None

        self.inputAction  = None
        self.outputAction = None

    def _setContent(self):
        '''create string representation of operation.
        '''
        kwstring = 'kw = {}'
        tCheck = 'if isinstance(request, %s) is False:' % self.inputName
        if self.wsA:
            wsactionIn = 'wsaction = "%s"' % self.inputAction
            wsactionOut = 'wsaction = "%s"' % self.outputAction
            bindArgs = 'wsaction=wsaction, endPointReference=self.endPointReference, **kw)'
            responseArgs = ', wsaction=wsaction'
        else:
            wsactionIn = '# no input wsaction'
            wsactionOut = '# no output wsaction'
            bindArgs = '**kw)'
            responseArgs = ''

        method = [
            '%s# op: %s' % (ID1, self.name),
            '%sdef %s(self, request):' % (ID1, self.name),
            '%s%s' % (ID2, tCheck),
            '%sraise TypeError, "%%s incorrect request type" %% (%s)' %(ID3, 'request.__class__'),
            '%s%s' % (ID2, kwstring),
            '%s%s' % (ID2, wsactionIn),
            '%sself.binding.Send(None, None, request, soapaction="%s", %s' %(ID2, self.soapaction, bindArgs),
            ]
        if self.outputName:
            response = ['%s%s' % (ID2, wsactionOut),]
            response.append('%sresponse = self.binding.Receive(%s.typecode%s)' %(ID2, self.outputName, responseArgs))
            if self.outputSimpleType:
                response.append('%sreturn %s(response)' %(ID2, self.outputName))
            else: 
                rCheck = 'if isinstance(response, %s.typecode.pyclass) is False:' %self.outputName
                response.append('%s%s' %(ID2, rCheck))
                response.append('%sraise TypeError, "%%s incorrect response type" %% (%s)' %(ID3, 'response.__class__'))
                response.append('%sreturn response' %ID2)
            method += response

        self.writeArray(method)


class ServiceMessageContainerBase(ServiceContainerBase):
    def __init__(self):
        ServiceContainerBase.__init__(self)
        
        self.pName = None
        self.mName = None


class ServiceDocumentLiteralMessageContainer(ServiceMessageContainerBase):
    def __init__(self, simple=None):
        '''
        simple -- primitive type message represents, or None if not primitive.
        '''
        ServiceMessageContainerBase.__init__(self)
        self._simple = simple

    def _setContent(self):
        '''create string representation of doc/lit message container.  If 
        message element is simple(primitive), use python type as base class.
        '''
        message = []
        if self._simple is None:
            message.append('class %s:' %self.mName)
        else:
            message.append('class %s(%s):' %(self.mName, self._simple))

        nsa = self.getNSAlias()
        message.append('%stypecode = %s.%s_Dec()' %(ID1, nsa, self.pName))
        self.writeArray(message)


class ServiceRPCLiteralMessageContainer(ServiceMessageContainerBase):
    '''
    '''
    def __init__(self, operation, boperation, input=False):
        '''Instance Data: 
               operation -- WSDLTools Operation instance
               boperation -- WSDLTools BindingOperation instance
               input -- boolean input/output
        '''
        ServiceMessageContainerBase.__init__(self)
        self.operation = operation
        self.boperation = boperation
        self.input = input

    def _setContent(self):
        operation = self.operation
        input = self.input
        pname = operation.name
        msgRole = operation.input
        msgRoleB = self.boperation.input
        if input is False:
            pname = '%sResponse' %operation.name
            msgRole = operation.output
            msgRoleB = self.boperation.output

        sbody = msgRoleB.findBinding(SoapBodyBinding)
        if not sbody or not sbody.namespace:            
            raise WSInteropError, WSISpec.R2717

        mName = msgRole.getMessage().name
        parts = msgRole.getMessage().parts.data.values()
        tcb = MessageTypecodeContainer(tuple(parts))
        ofwhat = '[%s]' %tcb.getTypecodeList()
        anames = tcb.getAttributeNames()
        pyclass = mName

        message = [
            'class %s:' %mName,
            '%sdef __init__(self):' % ID1,
            ]
        for aname in anames:
            message.append('%sself.%s = None' %(ID2, aname))
        if not len(anames):
            message.append('%sreturn' %ID2)

        message.append('%s.typecode = Struct(pname=("%s","%s"), ofwhat=%s, pyclass=%s)' \
            %(mName, sbody.namespace, pname, ofwhat, pyclass),)
        self.writeArray(message)


class ServiceEncodedMessageContainer(ServiceMessageContainerBase):

    def _setContent(self):

        message = [
            '# bogus encoded message',
            ]

        self.writeArray(message)

# -- containers for types file components

class TypesContainerBase(ContainerBase):
    def __init__(self):
        ContainerBase.__init__(self)


class TypesHeaderContainer(TypesContainerBase):
        
    def _setContent(self):
        
        imports = [
            'import ZSI',
            'import ZSI.TCcompound',
            'from ZSI.generate.utility import ElementDeclaration,TypeDefinition',
        ]

        self.writeArray(imports)

class NamespaceClassContainerBase(TypesContainerBase):
    def __init__(self):
        TypesContainerBase.__init__(self)
        

class NamespaceClassHeaderContainer(NamespaceClassContainerBase):

    def _setContent(self):

        head = [
            '#' * 30,
            '# targetNamespace',
            '# %s' % self.ns,
            '#' * 30 + '\n',
            'class %s:' % self.getNSAlias(),
            '%stargetNamespace = "%s"' % (ID1, self.ns)
            ]

        self.writeArray(head)

class NamespaceClassFooterContainer(NamespaceClassContainerBase):

    def _setContent(self):

        foot = [
            '# end class %s (tns: %s)' % (self.getNSAlias(), self.ns),
            ]

        self.writeArray(foot)

# -- containers for acutal typecode components

class TypecodeContainerBase(TypesContainerBase):
    '''Base class for all classes representing anything
    with element content.
    '''
    
    def __init__(self):
        TypesContainerBase.__init__(self)
        self.name = None
        self.bti = BaseTypeInterpreter()

        # attrs for model groups and others with elements, tclists, etc...
        self.allOptional = False
        self.mgContent = None
        self.contentFlattened = False
        self.elementAttrs = []
        self.elementsSet = False
        self.tcListElements = []
        self.tcListSet = False

        self.subTypes = []

        # used when processing nested anonymous types
        self.parentClass = None

        # used when processing attribute content

        self.mixed = False
        self.extraFlags = ''

        self.attributeContent = None
        self.attrComponents = []

    # following methods used to propogate typecode 'components' to
    # the inhereting container classes

    def getClassName(self):

        if not self.name:
            raise ContainerError, 'self.name not defined!'
        if not hasattr(self.__class__, 'type'):
            raise ContainerError, 'container type not defined!'

        suffix = ''

        if self.__class__.type == 'def':
            suffix = '_Def'
        elif self.__class__.type == 'dec':
            suffix = '_Dec'
        else:
            raise ContainerError, 'unknown container instance type: %s' \
                  % self.__class__.type
        
        return self.mangle( self.name + suffix )


    def getPyClass(self):
        '''Name of generated inner class that will be specified as pyclass.
        '''
        return 'Holder'

    def getPyClassDefinition(self):
        '''Return a list containing pyclass definition.
        '''
        definition = []
        definition.append('%sclass %s:' %(ID3,self.getPyClass()))
        definition.append('%stypecode = self' %ID4)
        definition.append('%sdef __init__(self):' %ID4)
        definition.append('%s# pyclass' %ID5)

        # JRB HACK need to call _setElements via getElements
        self.getElements()
        # JRB HACK need to indent additional one level
        for el in self.elementAttrs:
            definition.append('%s%s' %(ID2,el))
        definition.append('%sreturn' %ID5)

        return definition

    def nsuriLogic(self):
        '''set a variable "ns" that represents the targetNamespace in
        which this item is defined.  Used for namespacing local elements.
        '''
        if self.parentClass:
            return 'ns = %s.%s.schema' %(self.parentClass, self.getClassName())
        return 'ns = %s.%s.schema' %(self.getNSAlias(), self.getClassName())

    def schemaTag(self):
        return 'schema = "%s"' % self.ns
    
    def typeTag(self):
        return 'type = (schema, "%s")' % self.name
    
    def literalTag(self):
        return 'literal = "%s"' % self.name

    #def pname(self):
    #    return 'kw["pname"] = self.get_pname()'

    #def aname(self):
    #    return 'kw["aname"] = self.get_aname(kw["pname"][1])'

    def getExtraFlags(self):
        if self.mixed:
            self.extraFlags += 'mixed=True, mixed_aname="_text", '

        return self.extraFlags

    def simpleConstructor(self, superclass=None):

        if superclass:
            return '%s.__init__(self, **kw)' % superclass
        else:
            return 'def __init__(self, **kw):'

    def pnameConstructor(self, superclass=None):

        if superclass:
            return '%s.__init__(self, pname, **kw)' % superclass
        else:
            return 'def __init__(self, pname, **kw):'

    # _flattenContent() is a pre-processor called before either the
    # elements or typecode strings are generated.  it flattens out
    # embedded/anonymous model group elements.  _setElements() and
    # _setTypecodeList() skip the model groups and pay attention
    # to the elements contained therein.

    def _flattenContent(self):
        if not self.contentFlattened:
            self.contentFlattened = True
            for c in self.mgContent:
                if c.isChoice() or c.isSequence():
                    self.mgContent += c.content

    # getElements() and _setElements() generate the class variable
    # elements lists in definition classes

    def _setElements(self):

        if type(self.mgContent) is not tuple:
            raise TypeError, 'expecting tuple for mgContent got %s' %type(self.mgContent)

        for c in self.mgContent:
            if c.isDeclaration() and c.isElement():
                if c.getAttribute('name'):
                    e = '%sself._%s = None' \
                        % (ID3, self.mangle(c.getAttribute('name')))
                    self.elementAttrs.append(e)
                elif c.isWildCard():
                    e = '%sself._any = None' % (ID3)
                else:
                    raise ContainerError, 'ouch'
            elif c.isReference():
                if c.getAttribute('ref')[1]:
                    e = '%sself._%s = None' \
                        % (ID3,self.mangle(c.getAttribute('ref')[1]))
                    self.elementAttrs.append(e)
                else:
                    raise ContainerError, 'ouch as well'
            elif c.isSequence() or c.isChoice():
                if not self.contentFlattened:
                    raise ContainerError, 'unflattened content found!'
                else:
                    # model group contents already flattened
                    pass
            else:
                raise ContainerError, 'unexpected item: %s' % c.getItemTrace()

    def getElements(self):
        if not self.elementsSet:
            self._flattenContent()
            self._setElements()
            self.elementsSet = True

        return '\n'.join(self.elementAttrs)

    def _setTypecodeList(self):
        '''getTypecodeList and _setTypecodeList generates the TClist 
        strings where appropriate.
        '''
        if type(self.mgContent) is not tuple:
            raise TypeError, 'expecting tuple for mgContent got %s' %type(self.mgContent)

        for c in self.mgContent:
            tc = TcListComponentContainer()
            min,max,nil = self._getOccurs(c)
            tc.setOccurs(min, max, nil)
            if c.isDeclaration() and c.isElement():
                if c.isLocal() and c.isQualified() is False:
                    tc.unQualified()
                if c.isWildCard():
                    tc.setStyleAnyElement()
                elif self._elementContainsLocalDef(c):
                    # element containing a local def
                    tc.name = c.getAttribute('name')
                    tc.klass = 'self.__class__.%s_Dec' % tc.name
                    tc.setStyleElementReference()
                    # add class subtype for processing
                    self.subTypes.append(c)
                elif c.getAttribute('type'):
                    tc.name = c.getAttribute('name')
                    if self._isXSD(c):
                        # an xsd primitive
                        tpc = self.bti.get_typeclass(c.getAttribute('type')[1],
                                                     c.getAttribute('type')[0])
                        tc.klass = tpc
                    elif self._isRecursiveElement(c):
                        tc.setStyleRecursion()
                    else:
                        # a definition
                        ns = c.getAttribute('type')[0]
                        tc.klass = '%s.%s_Def' % (self.nsad.getAlias(ns),
                                                  c.getAttribute('type')[1])
                else:
                    raise ContainerError, 'unexpected item: %s' % c.getItemTrace()
            elif c.isReference():
                # element references
                ns = c.getAttribute('ref')[0]
                tc.klass = '%s.%s_Dec' % (self.nsad.getAlias(ns),
                                          c.getAttribute('ref')[1])
                tc.setStyleElementReference()
            elif c.isSequence() or c.isChoice():
                if not self.contentFlattened:
                    raise ContainerError, 'unflattened content found!'
                continue
            else:
                raise ContainerError, 'unexpected item: %s' % c.getItemTrace()

            self.tcListElements.append(tc)

    def getTypecodeList(self):
        if not self.tcListSet:
            self._flattenContent()
            self._setTypecodeList()
            self.tcListSet = True

        list = []
        for e in self.tcListElements:
            list.append(str(e))

        return ', '.join(list)

    # the following _methods() are utility methods used during
    # TCList generation, et al.
    
    def _elementContainsLocalDef(self, c):
        if c.getTypeDefinition('type') == None:
            if hasattr(c, 'content'):
                if isinstance(c.content, XMLSchema.LocalComplexType):
                    return True
                
        return False

    def _isXSD(self, c):
        if c.getAttribute('type'):
            if c.getAttribute('type')[0] in SCHEMA.XSD_LIST:
                return True
            
        return False

    def _isRecursiveElement(self, c):
        if self.name == c.getAttribute('type')[1]:
            if self.ns == c.getAttribute('type')[0]:
                return True

        return False

    def _getOccurs(self, e):
        
        nillable = e.getAttribute('nillable')
        if nillable == 'true':
            nillable = '1'
        else:
            nillable = '0'
            
        maxOccurs = e.getAttribute('maxOccurs')
        if maxOccurs == 'unbounded':
            maxOccurs = '"%s"' %maxOccurs
            
        minOccurs = e.getAttribute('minOccurs')
        
        if self.allOptional is True:
            #JRB Hack
            minOccurs = '0'
            maxOccurs = '"unbounded"'

        return minOccurs,maxOccurs,nillable

    # these methods are used when generating attribute content:

    def _setAttributes(self):
        '''set up generated typecode.
        '''
        if not self.attributeContent:
            return
        self.attrComponents.append('# attribute handling code')
        self.attrComponents.append('self.attribute_typecode_dict.update(%s.%s.attribute_typecode_dict)' \
            %(self.getNSAlias(), self.getClassName()))

        for a in self.attributeContent:
            if a.isWildCard() and a.isAttribute() and a.isDeclaration():
                self.attrComponents.append('self.attribute_typecode_dict[("%s","anyAttribute")] = ZSI.TC.AnyElement()' % (SCHEMA.XSD3))
            elif a.isDeclaration():
                t = a.getAttribute('type')
                try:
                    tc = self.bti.get_typeclass(t[1], t[0])
                except:
                    raise ContainerError, 'choked on this attr: %s', t

                key = None
                if a.getAttribute('form') == 'qualified':
                    key = '("%s","%s")' % ( a.getTargetNamespace(),
                                            a.getAttribute('name') )
                elif a.getAttribute('form') == 'unqualified':
                    key = '"%s"' % a.getAttribute('name')
                else:
                    raise ContainerError, \
                          'attribute form must be un/qualified %s' \
                          % a.getAttribute('form')
                self.attrComponents.append('self.attribute_typecode_dict[%s] = %s()' % (key, tc))
                
            elif a.isReference():
                if a.isAttributeGroup():
                    # flatten 'em out....
                    for ga in a.getAttributeGroup().getAttributeContent():
                        if not ga.isAttributeGroup():
                            self.attributeContent += (ga,)
                    continue
                ga = a.getAttributeDeclaration()
                tp = ga.getTypeDefinition('type')
                if tp is None:
                    # built in simple type
                    try:
                        namespace,typeName = ga.getAttribute('type')
                    except TypeError, ex:
                        raise ContainerError,\
                              'no type in attribute declaration!'
                    tc = self.bti.get_typeclass(typeName, namespace)
                    key = '("%s","%s")' \
                          % (ga.getTargetNamespace(),ga.getAttribute('name'))
                    self.attrComponents.append('self.attribute_typecode_dict[%s] = %s()' % (key, tc))
                else:
                    typeName = tp.getAttribute('name')
                    namespace = tp.getTargetNamespace()
                    alias = self.nsad.getAlias(namespace)
                    key = '("%s","%s")' \
                          % (ga.getTargetNamespace(),ga.getAttribute('name'))
                    self.attrComponents.append('self.attribute_typecode_dict[%s] = %s.%s_Def(None)' % (key, alias, typeName))
                
            else:
                raise ContainerError, 'unknown attribute type!'

        #self.attrComponents.append('setattr(self, self.attrs_aname, {})')

    def getAttributes(self):
        return self.attrComponents

    def getBasesLogic(self, indent):

        retval = ''

        bases = [
            'if %s.%s_Def not in %s.%s.__bases__:' \
            % (self.nsad.getAlias(self.sKlassNS), self.sKlass,
               self.getNSAlias(), self.getClassName()),
            '%sbases = list(%s.%s.__bases__)' \
            % (ID1,self.getNSAlias(),self.getClassName()),
            '%sbases.insert(0, %s.%s_Def)' \
            % (ID1,self.nsad.getAlias(self.sKlassNS), self.sKlass),
            '%s%s.%s.__bases__ = tuple(bases)' \
            % (ID1, self.getNSAlias(), self.getClassName())
            ]

        for b in bases:
            retval += '%s%s\n' % (indent, b)

        return retval

class MessageTypecodeContainer(TypecodeContainerBase):
    '''Used for RPC style messages, where we have 
    serveral parts serialized within a rpc wrapper name.
    '''
    def __init__(self, parts=None):
        TypecodeContainerBase.__init__(self)
        self.mgContent = parts

    def _getOccurs(self, e):
        '''return a 3 item tuple 
        '''
        minOccurs = maxOccurs = nillable = '1'
        return minOccurs,maxOccurs,nillable

    def _setTypecodeList(self):
        if type(self.mgContent) is not tuple:
            raise TypeError, 'expecting tuple for mgContent got %s' %type(self.mgContent)

        for p in self.mgContent:
            # JRB
            #   not sure what needs to be set for tc, this should be
            #   the job of the constructor or a setUp method.
            min,max,nil = self._getOccurs(p)
            if p.element:
                raise  WSInteropError, WSISpec.R2203
            elif p.type: 
                nsuri,name = p.type
                tc = TcListComponentContainer(qualified=False)
                tc.setOccurs(min, max, nil)
                tc.name = p.name
                if nsuri in SCHEMA.XSD_LIST:
                    tpc = self.bti.get_typeclass(name, nsuri)
                    tc.klass = tpc
                else:
                    tc.klass = '%s.%s_Def' % (self.nsad.getAlias(nsuri), name)
            else:
                raise ContainerError, 'part must define an element or type attribute'

            self.tcListElements.append(tc)

    def getTypecodeList(self):
        if not self.tcListSet:
            self._setTypecodeList()
            self.tcListSet = True

        list = []
        for e in self.tcListElements:
            list.append(str(e))
        return ', '.join(list)

    def getAttributeNames(self):
        '''returns a list of anames representing the parts
        of the message.
        '''
        return map(lambda e: e.getAttributeName(), self.tcListElements)

    def setParts(self, parts):
        self.mgContent = parts


class TcListComponentContainer(TypecodeContainerBase):
    '''Encapsulates a single value in the TClist list.
    it inherits TypecodeContainerBase only to get the mangle() method,
    it does not call the baseclass ctor.
    '''
    
    def __init__(self, qualified=True):
        '''
        qualified -- qualify element.  All GEDs should be qualified,
            but local element declarations qualified if form attribute
            is qualified, else they are unqualified. Only relevant for
            standard style.
        '''
        self.qualified = qualified
        self.name = None
        self.klass = None
        
        self.min = None
        self.max = None
        self.nil = None
        self.setStyleElementDeclaration()

    def setOccurs(self, min, max, nil):
        self.min = min
        self.max = max
        self.nil = nil

    def setStyleElementDeclaration(self):
        '''set the element style.
            standard -- GED or local element
        '''
        self.style = 'standard'

    def setStyleElementReference(self):
        '''set the element style.
            ref -- element reference
        '''
        self.style = 'ref'

    def setStyleAnyElement(self):
        '''set the element style.
            anyElement -- <any> element wildcard
        '''
        self.name = 'any'
        self.style = 'anyElement'

    def setStyleRecursion(self):
        '''set the element style.
            recursion -- do lazy evaluation of element/type because
               the element contains an instance of itself.
        '''
        self.style = 'recursion'

    def unQualified(self):
        '''Do not qualify element.
        '''
        self.qualified = False

    def getAttributeName(self, name=None):
        '''represents the aname
        '''
        name = name or self.name
        return '_%s' %self.mangle(name)

    def _getOccurs(self):
        return 'minOccurs=%s, maxOccurs=%s, nillable=%s' \
               % (self.min, self.max, self.nil)

    def _getvalue(self):
        if self.style == 'standard':
            pname = '"%s"' %self.name
            if self.qualified is True:
                pname = '(ns,"%s")' %self.name
            return '%s(pname=%s, aname="%s", %s)' \
                   % (self.klass, pname,
                      self.getAttributeName(), self._getOccurs())
        elif self.style == 'ref':
            return '%s(%s)' % (self.klass, self._getOccurs())
        elif self.style == 'anyElement':
            return 'ZSI.TC.AnyElement(aname="%s", %s)' \
                %(self.getAttributeName(), self._getOccurs())
        elif self.style == 'recursion':
            # JRB why is aname w/o _?
            return 'ZSI.TC.AnyElement(aname="%s", %s)' \
                % (self.getAttributeName(), self._getOccurs())

    def __str__(self):
        return self._getvalue()
    

# The following methods contain the actual typecode variation containers.
# the formatting ought to be fairly skeletal in each one, relying on
# base class methods to propogate a good deal of the logic components
# to them.

# placeholder for testing
class GenericDefinitionTypecodeContainer(TypecodeContainerBase):

    type = 'def'

    def _setContent(self):

        type = [
            '%sclass %s(ZSI.TCcompound.Struct):' % (ID1,self.getClassName()),
            '%s# definition' % ID2,
            '%spass' % ID2,
            ]

        self.writeArray(type)

# placeholder for testing
class GenericElementTypecodeContainer(TypecodeContainerBase):

    type = 'dec'

    def _setContent(self):

        type = [
            '%sclass %s:' % (ID1,self.getClassName()),
            '%s# element' % ID2,
            '%spass' % ID2,
            ]

        self.writeArray(type)

class ElementSimpleTypeContainer(TypecodeContainerBase):

    type = 'dec'

    def _setContent(self):

        element = [
            '%sclass %s(%s, ElementDeclaration):' % (ID1, self.getClassName(),
                                                     self.sKlass),
            '%s%s' % (ID2, self.literalTag()),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.simpleConstructor()),
            '%skw["pname"] = ("%s","%s")' % (ID3, self.ns, self.name),
            '%skw["aname"] = "_%s"' % (ID3, self.name),
            '%s%s' % (ID3, self.simpleConstructor(self.sKlass)),
            ]

        self.writeArray(element)

class ElementLocalDefContainer(TypecodeContainerBase):

    type = 'dec'

    def _setContent(self):

        element = [
            '%sclass %s(ZSI.TCcompound.Struct, ElementDeclaration):' \
            % (ID1,self.getClassName()),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.literalTag()),
            '%s%s' % (ID2, self.simpleConstructor()),
            #'%s' % self.getElements(),
            '%s%s' % (ID3, self.nsuriLogic()),
            '%sTClist = [%s]' % (ID3, self.getTypecodeList()),
            '%skw["pname"] = ("%s","%s")' % (ID3, self.ns, self.name),
            '%skw["aname"] = "_%s"' % (ID3, self.name),
            ]

        # Write out a pyclass
        for l in self.getPyClassDefinition():
            element.append(l)

        element.append(\
            '%sZSI.TCcompound.Struct.__init__(self, %s, TClist, inorder=0, **kw)' \
            %(ID3, self.getPyClass())
        )
        self.writeArray(element)

class ElementGlobalDefContainer(TypecodeContainerBase):

    type = 'dec'

    def _setContent(self):
        '''GED defines element name, so also define typecode aname
        '''
        element = [
            '%sclass %s(ElementDeclaration):' % (ID1, self.getClassName()),
            '%s%s' % (ID2, self.literalTag()),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.simpleConstructor()),
            '%skw["pname"] = ("%s","%s")' % (ID3, self.ns, self.name),
            '%skw["aname"] = "_%s"' % (ID3, self.name),
            '%s'   % self.getBasesLogic(ID3),
            '%s%s.%s_Def.__init__(self, **kw)' \
            % (ID3, self.nsad.getAlias(self.sKlassNS), self.sKlass),
            ]

        self.writeArray(element)

class DerivedTypeDefContainer(TypecodeContainerBase):

    type = 'def'

    def __init__(self):
        TypecodeContainerBase.__init__(self)

        self.restriction = False
        self.extension = False

        # this is used in figuring attrs for extensions
        self.extType = None

    def _setContent(self):

        if self.restriction:
            assigner = '='
        else:
            assigner = '+='

        definition = [
            '%sclass %s(TypeDefinition):' % (ID1, self.getClassName()),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.typeTag()),
            '%sattribute_typecode_dict = {}' % ID2,
            '%s%s' % (ID2, self.pnameConstructor()),
            '%s%s' % (ID3, self.nsuriLogic()),
            '%sTClist = [%s]' % (ID3, self.getTypecodeList()),
            '%s'   % self.getBasesLogic(ID3),
            '%s%s.%s_Def.__init__(self, pname, **kw)' \
            % (ID3, self.nsad.getAlias(self.sKlassNS), self.sKlass),
            '%sself.ofwhat %s tuple(TClist)' %(ID3, assigner),
            ]

        self._setAttributes()
        for l in self.getAttributes():
            definition.append('%s%s'%(ID3, l))

        self.writeArray(definition)

class ModelGroupContainer(TypecodeContainerBase):

    type = 'def'

    def _setContent(self):

        definition = [
            '%sclass %s(ZSI.TCcompound.Struct, TypeDefinition):'
            % (ID1, self.getClassName()),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.typeTag()),
            '%sattribute_typecode_dict = {}' % ID2,
            '%s%s' % (ID2, self.pnameConstructor()),
            #'%s'   % self.getElements(),
            '%s%s' % (ID3, self.nsuriLogic()),
            '%sTClist = [%s]' % (ID3, self.getTypecodeList()),
            ]

        self._setAttributes()
        for l in self.getAttributes():
            definition.append('%s%s'%(ID3, l))

        # write out a pyclass
        for l in self.getPyClassDefinition():
            definition.append(l)

        definition.append(\
            '%sZSI.TCcompound.Struct.__init__(self, %s, TClist, pname=pname, inorder=0, %s**kw)' \
            %(ID3, self.getPyClass(), self.getExtraFlags())
        )

        self.writeArray(definition)

class SimpleTypeContainer(TypecodeContainerBase):

    type = 'def'

    def __init__(self):
        TypecodeContainerBase.__init__(self)

    def setUp(self, name, tns, pyclass):
        '''
        name -- type definition name
        tns -- targetNamespace type is defined in.
        pyclass -- class representing base type.
        '''
        self.name,self.ns,self.sKlass = name, tns, pyclass

    def _setContent(self):

        definition = [
            '%sclass %s(%s, TypeDefinition):' \
            % (ID1, self.getClassName(), self.sKlass),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.typeTag()),
            '%s%s' % (ID2, self.pnameConstructor()),
            '%s%s' % (ID3, self.pnameConstructor(self.sKlass)),
            
            ]

        self.writeArray(definition)

class SimpleContentContainer(TypecodeContainerBase):

    type = 'def'

    def _setContent(self):

        definition = [
            '%sclass %s(%s, TypeDefinition):' \
            % (ID1, self.getClassName(), self.sKlass),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.typeTag()),
            '%sattribute_typecode_dict = {}' % ID2,
            '%s%s' % (ID2, self.pnameConstructor()),
            '%s%s' % (ID3, self.pnameConstructor(self.sKlass)),
            ]

        self._setAttributes()
        for l in self.getAttributes():
            definition.append('%s%s'%(ID3, l))
        self.writeArray(definition)

class DerivedSimpleContentContainer(TypecodeContainerBase):

    type = 'def'

    def _setContent(self):

        definition = [
            '%sclass %s(TypeDefinition):' % (ID1, self.getClassName()),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.typeTag()),
            '%sattribute_typecode_dict = {}' % ID2,
            '%s%s' % (ID2, self.pnameConstructor()),
            '%s%s' % (ID3, self.nsuriLogic()),
            '%s'   % self.getBasesLogic(ID3),
            '%s%s.%s_Def.__init__(self, pname, **kw)' \
            % (ID3, self.nsad.getAlias(self.sKlassNS), self.sKlass),
            ]

        self._setAttributes()
        for l in self.getAttributes():
            definition.append('%s%s'%(ID3, l))
        self.writeArray(definition)


class UnionOrListContainer(SimpleContentContainer):

    type = 'def'
